

{-# OPTIONS_GHC -optc-D__HUGS__ #-}
{-# INCLUDE <libpq-fe.h> #-}
{-# INCLUDE <pg_config.h> #-}
{-# LINE 1 "Connection.hsc" #-}
-- -*- mode: haskell; -*-
{-# LINE 2 "Connection.hsc" #-}
{-# CFILES hugs-postgresql-helper.c #-}
-- Above line for hugs
{-
Copyright (C) 2005 John Goerzen <jgoerzen@complete.org>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
-}

module Database.HDBC.PostgreSQL.Connection (connectPostgreSQL) where

import Database.HDBC.Types
import Database.HDBC
import Database.HDBC.PostgreSQL.Types
import Database.HDBC.PostgreSQL.Statement
import Foreign.C.Types
import Foreign.C.String
import Foreign.Marshal
import Foreign.Storable
import Database.HDBC.PostgreSQL.Utils
import Foreign.ForeignPtr
import Foreign.Ptr
import Data.Word


{-# LINE 38 "Connection.hsc" #-}

{-# LINE 39 "Connection.hsc" #-}

{- | Connect to a PostgreSQL server.

See <http://www.postgresql.org/docs/8.1/static/libpq.html#LIBPQ-CONNECT> for the meaning
of the connection string. -}
connectPostgreSQL :: String -> IO Connection
connectPostgreSQL args = withCString args $
  \cs -> do ptr <- pqconnectdb cs
            status <- pqstatus ptr

{-# LINE 49 "Connection.hsc" #-}
            env <- pqhugs_alloc_env
            fptr <- newForeignPtrEnv pqfinishptr env ptr

{-# LINE 54 "Connection.hsc" #-}

            case status of

{-# LINE 57 "Connection.hsc" #-}
                     0 -> mkConn args env fptr
{-# LINE 58 "Connection.hsc" #-}

{-# LINE 61 "Connection.hsc" #-}
                     _ -> raiseError "connectPostgreSQL" status ptr

-- FIXME: environment vars may have changed, should use pgsql enquiries
-- for clone.

{-# LINE 66 "Connection.hsc" #-}
mkConn :: String -> Ptr CInt -> Conn -> IO Connection
mkConn args env conn = withConn conn $

{-# LINE 72 "Connection.hsc" #-}
  \cconn -> 
    do begin_transaction conn
       protover <- pqprotocolVersion cconn
       serverver <- pqserverVersion cconn
       let clientver = "8.1.3"
{-# LINE 77 "Connection.hsc" #-}
       return $ Connection {

{-# LINE 79 "Connection.hsc" #-}
                            disconnect = fdisconnect env conn,

{-# LINE 83 "Connection.hsc" #-}
                            commit = fcommit conn,
                            rollback = frollback conn,
                            run = frun conn,
                            prepare = newSth conn,
                            clone = connectPostgreSQL args,
                            hdbcDriverName = "postgresql",
                            hdbcClientVer = clientver,
                            proxiedClientName = "postgresql",
                            proxiedClientVer = show protover,
                            dbServerVer = show serverver,
                            getTables = fgetTables conn}

--------------------------------------------------
-- Guts here
--------------------------------------------------

begin_transaction :: Conn -> IO ()
begin_transaction o = frun o "BEGIN" [] >> return ()

frun o query args =
    do sth <- newSth o query
       res <- execute sth args
       finish sth
       return res

fcommit o = do frun o "COMMIT" []
               begin_transaction o
frollback o =  do frun o "ROLLBACK" []
                  begin_transaction o

fgetTables conn =
    do sth <- newSth conn "select table_name from information_schema.tables where table_schema = 'public'"
       execute sth []
       res1 <- fetchAllRows sth
       let res = map fromSql $ concat res1
       return $ seq (length res) res


{-# LINE 121 "Connection.hsc" #-}
fdisconnect env conn = withConn conn (\cconn -> pqfinish env cconn)

{-# LINE 125 "Connection.hsc" #-}

foreign import ccall unsafe "libpq-fe.h PQconnectdb"
  pqconnectdb :: CString -> IO (Ptr CConn)

foreign import ccall unsafe "libpq-fe.h PQstatus"
  pqstatus :: Ptr CConn -> IO Word32
{-# LINE 131 "Connection.hsc" #-}


{-# LINE 133 "Connection.hsc" #-}
-- already imported from Statement.hsc
--foreign import ccall unsafe "hugs-postgresql-helper.h PQhugs_alloc_env"
--  pqhugs_alloc_env :: IO (Ptr CInt)

foreign import ccall unsafe "hugs-postgresql-helper.h PQfinish_hugs_app"
  pqfinish :: Ptr CInt -> Ptr CConn -> IO ()

foreign import ccall unsafe "hugs-postgresql-helper.h &PQfinish_hugs_fptr"
  pqfinishptr :: FunPtr (Ptr CInt -> Ptr CConn -> IO ())

{-# LINE 146 "Connection.hsc" #-}

foreign import ccall unsafe "libpq-fe.h PQprotocolVersion"
  pqprotocolVersion :: Ptr CConn -> IO CInt

foreign import ccall unsafe "libpq-fe.h PQserverVersion"
  pqserverVersion :: Ptr CConn -> IO CInt
