

{-# OPTIONS_GHC -optc-D__HUGS__ #-}
{-# INCLUDE <libpq-fe.h> #-}
{-# LINE 1 "Statement.hsc" #-}
-- -*- mode: haskell; -*-
{-# LINE 2 "Statement.hsc" #-}
{-# CFILES hugs-postgresql-helper.c #-}
-- Above line for hugs
{-
Copyright (C) 2005 John Goerzen <jgoerzen@complete.org>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
-}
module Database.HDBC.PostgreSQL.Statement where
import Database.HDBC.Types
import Database.HDBC
import Database.HDBC.PostgreSQL.Types
import Database.HDBC.PostgreSQL.Utils
import Foreign.C.Types
import Foreign.ForeignPtr
import Foreign.Ptr
import Control.Concurrent.MVar
import Foreign.C.String
import Foreign.Marshal
import Foreign.Storable
import Control.Monad
import Data.List
import Data.Word
import Control.Exception
import System.IO
import Database.HDBC.PostgreSQL.Parser(convertSQL)

l _ = return ()
--l m = hPutStrLn stderr ("\n" ++ m)


{-# LINE 44 "Statement.hsc" #-}

data SState = 
    SState { stomv :: MVar (Maybe Stmt),

{-# LINE 48 "Statement.hsc" #-}
             envmv :: MVar (Maybe (Ptr CInt)),

{-# LINE 50 "Statement.hsc" #-}
             nextrowmv :: MVar (CInt), -- -1 for no next row (empty); otherwise, next row to read.
             dbo :: Conn,
             squery :: String,
             colnamemv :: MVar [String]}

-- FIXME: we currently do no prepare optimization whatsoever.

newSth :: Conn -> String -> IO Statement               
newSth indbo query = 
    do l "in newSth"
       newstomv <- newMVar Nothing

{-# LINE 62 "Statement.hsc" #-}
       newenvmv <- newMVar Nothing

{-# LINE 64 "Statement.hsc" #-}
       newnextrowmv <- newMVar (-1)
       newcolnamemv <- newMVar []
       usequery <- case convertSQL query of
                      Left errstr -> throwDyn $ SqlError
                                      {seState = "",
                                       seNativeError = (-1),
                                       seErrorMsg = "hdbc prepare: " ++ 
                                                    show errstr}
                      Right converted -> return converted
       let sstate = SState {stomv = newstomv, nextrowmv = newnextrowmv,
                            dbo = indbo, squery = usequery,

{-# LINE 76 "Statement.hsc" #-}
                            envmv = newenvmv,

{-# LINE 78 "Statement.hsc" #-}
                            colnamemv = newcolnamemv}
       return $ Statement {execute = fexecute sstate,
                           executeMany = fexecutemany sstate,
                           finish = public_ffinish sstate,
                           fetchRow = ffetchrow sstate,
                           originalQuery = query,
                           getColumnNames = readMVar (colnamemv sstate)}

{- For now, we try to just  handle things as simply as possible.
FIXME lots of room for improvement here (types, etc). -}
fexecute sstate args = withForeignPtr (dbo sstate) $ \cconn ->
                       withCString (squery sstate) $ \cquery ->
                       withCStringArr0 args $ \cargs ->
    do l "in fexecute"
       public_ffinish sstate    -- Sets nextrowmv to -1
       resptr <- pqexecParams cconn cquery
                 (genericLength args) nullPtr cargs nullPtr nullPtr 0
       status <- pqresultStatus resptr
       case status of
         0 ->
{-# LINE 98 "Statement.hsc" #-}
             do l $ "PGRES_EMPTY_QUERY: " ++ squery sstate
                pqclear resptr
                swapMVar (colnamemv sstate) []
                return 0
         1 ->
{-# LINE 103 "Statement.hsc" #-}
             do l $ "PGRES_COMMAND_OK: " ++ squery sstate
                rowscs <- pqcmdTuples resptr
                rows <- peekCString rowscs
                pqclear resptr
                swapMVar (colnamemv sstate) []
                return $ case rows of
                                   "" -> 0
                                   x -> read x
         2 -> 
{-# LINE 112 "Statement.hsc" #-}
             do l $ "PGRES_TUPLES_OK: " ++ squery sstate
                fgetcolnames resptr >>= swapMVar (colnamemv sstate) 
                numrows <- pqntuples resptr
                if numrows < 1
                   then do pqclear resptr
                           return 0
                   else do 

{-# LINE 120 "Statement.hsc" #-}
                        env <- pqhugs_alloc_env
                        fresptr <- newForeignPtrEnv pqclearptr env resptr

{-# LINE 125 "Statement.hsc" #-}
                        swapMVar (nextrowmv sstate) 0
                        swapMVar (stomv sstate) (Just fresptr)

{-# LINE 128 "Statement.hsc" #-}
                        swapMVar (envmv sstate) (Just env)

{-# LINE 130 "Statement.hsc" #-}
                        return 0
         _ -> do l $ "PGRES ERROR: " ++ squery sstate
                 csstatusmsg <- pqresStatus status
                 cserrormsg <- pqresultErrorMessage resptr
                 statusmsg <- peekCString csstatusmsg
                 errormsg <- peekCString cserrormsg
                 pqclear resptr
                 throwDyn $ 
                          SqlError {seState = "",
                                    seNativeError = fromIntegral status,
                                    seErrorMsg = "execute: " ++ statusmsg ++
                                                 ": " ++ errormsg}
{- General algorithm: find out how many columns we have, check the type
of each to see if it's NULL.  If it's not, fetch it as text and return that.
-}

ffetchrow :: SState -> IO (Maybe [SqlValue])
ffetchrow sstate = modifyMVar (nextrowmv sstate) dofetchrow
    where dofetchrow (-1) = l "ffr -1" >> return ((-1), Nothing)
          dofetchrow nextrow = modifyMVar (stomv sstate) $ \stmt -> 
             case stmt of
               Nothing -> l "ffr nos" >> return (stmt, ((-1), Nothing))
               Just cmstmt -> withStmt cmstmt $ \cstmt ->
                 do l $ "ffetchrow: " ++ show nextrow
                    numrows <- pqntuples cstmt
                    l $ "numrows: " ++ show numrows
                    if nextrow >= numrows
                       then do l "no more rows"
                               -- Don't use public_ffinish here

{-# LINE 160 "Statement.hsc" #-}
                               ffinish (envmv sstate) cmstmt

{-# LINE 164 "Statement.hsc" #-}
                               return (Nothing, ((-1), Nothing))
                       else do l "getting stuff"
                               ncols <- pqnfields cstmt
                               res <- mapM (getCol cstmt nextrow) 
                                      [0..(ncols - 1)]
                               return (stmt, (nextrow + 1, Just res))
          getCol p row icol = 
             do isnull <- pqgetisnull p row icol
                if isnull /= 0
                   then return SqlNull
                   else do text <- pqgetvalue p row icol
                           s <- peekCString text
                           return (SqlString s)

fgetcolnames cstmt =
    do ncols <- pqnfields cstmt
       mapM (\i -> pqfname cstmt i >>= peekCString) [0..(ncols - 1)]

-- FIXME: needs a faster algorithm.
fexecutemany :: SState -> [[SqlValue]] -> IO ()
fexecutemany sstate arglist =
    mapM_ (fexecute sstate) arglist >> return ()

-- Finish and change state
public_ffinish sstate = 
    do l "public_ffinish"
       swapMVar (nextrowmv sstate) (-1)
       modifyMVar_ (stomv sstate) worker
    where worker Nothing = return Nothing

{-# LINE 194 "Statement.hsc" #-}
          worker (Just sth) = ffinish (envmv sstate) sth >> return Nothing

{-# LINE 198 "Statement.hsc" #-}


{-# LINE 200 "Statement.hsc" #-}
ffinish :: MVar (Maybe (Ptr CInt)) -> Stmt -> IO ()
ffinish envmv p = modifyMVar_ envmv $ \mcenv ->
    do case mcenv of
          Nothing -> fail "Internal error, HDBC PostgreSQL hugs: ffinish called when no cenv present"
          Just cenv -> do withStmt p $ \cstmt -> pqclear_hugs cenv cstmt
                          return Nothing

{-# LINE 210 "Statement.hsc" #-}

foreign import ccall unsafe "libpq-fe.h PQresultStatus"
  pqresultStatus :: (Ptr CStmt) -> IO Word32
{-# LINE 213 "Statement.hsc" #-}

foreign import ccall unsafe "libpq-fe.h PQexecParams"
  pqexecParams :: (Ptr CConn) -> CString -> CInt ->
                  (Ptr Word32) ->
{-# LINE 217 "Statement.hsc" #-}
                  (Ptr CString) ->
                  (Ptr CInt) ->
                  (Ptr CInt) ->
                  CInt ->
                  IO (Ptr CStmt)


{-# LINE 224 "Statement.hsc" #-}
foreign import ccall unsafe "hugs-postgresql-helper.h PQhugs_alloc_env"
  pqhugs_alloc_env :: IO (Ptr CInt)

foreign import ccall unsafe "hugs-postgresql-helper.h PQclear_hugs_app"
  pqclear_hugs :: Ptr CInt -> Ptr CStmt -> IO ()

foreign import ccall unsafe "hugs-postgresql-helper.h &PQclear_hugs_fptr"
  pqclearptr :: FunPtr (Ptr CInt -> Ptr CStmt -> IO ())

{-# LINE 236 "Statement.hsc" #-}

foreign import ccall unsafe "libpq-fe.h PQclear"
  pqclear :: Ptr CStmt -> IO ()

foreign import ccall unsafe "libpq-fe.h PQcmdTuples"
  pqcmdTuples :: Ptr CStmt -> IO CString
foreign import ccall unsafe "libpq-fe.h PQresStatus"
  pqresStatus :: Word32 -> IO CString
{-# LINE 244 "Statement.hsc" #-}

foreign import ccall unsafe "libpq-fe.h PQresultErrorMessage"
  pqresultErrorMessage :: (Ptr CStmt) -> IO CString

foreign import ccall unsafe "libpq-fe.h PQntuples"
  pqntuples :: Ptr CStmt -> IO CInt

foreign import ccall unsafe "libpq-fe.h PQnfields"
  pqnfields :: Ptr CStmt -> IO CInt

foreign import ccall unsafe "libpq-fe.h PQgetisnull"
  pqgetisnull :: Ptr CStmt -> CInt -> CInt -> IO CInt

foreign import ccall unsafe "libpq-fe.h PQgetvalue"
  pqgetvalue :: Ptr CStmt -> CInt -> CInt -> IO CString

foreign import ccall unsafe "libpq-fe.h PQfname"
  pqfname :: Ptr CStmt -> CInt -> IO CString
